%% ``The contents of this file are subject to the Erlang Public License,
%% Version 1.1, (the "License"); you may not use this file except in
%% compliance with the License. You should have received a copy of the
%% Erlang Public License along with this software. If not, it can be
%% retrieved via the world wide web at http://www.erlang.org/.
%% 
%% Software distributed under the License is distributed on an "AS IS"
%% basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See
%% the License for the specific language governing rights and limitations
%% under the License.
%% 
%% The Initial Developer of the Original Code is Ericsson Utvecklings AB.
%% Portions created by Ericsson are Copyright 1999, Ericsson Utvecklings
%% AB. All Rights Reserved.''
%% 
%%     $Id$
%%
-module(win32reg_SUITE).

-export([all/1, erlang_entries/1, not_run/1, long/1, evil_write/1]).

-include("test_server.hrl").

all(suite) ->
    case os:type() of
	{win32, _} ->
	    [erlang_entries, long, evil_write];
	_ ->
	    [not_run]
    end.


not_run(doc) ->
    ["Doesn't run on UNIX"];
not_run(suite) -> [];
not_run(Config) when list(Config) ->
    {comment, "Doesn't run on UNIX."}.

%% Checks the read functions by reading the information about
%% Erlang which should be in the registry.

erlang_entries(suite) -> [];
erlang_entries(Config) when list(Config) ->
    ?line Dog = test_server:timetrap(test_server:seconds(10)),

    ?line {ok, Reg} = win32reg:open([read]),
    ?line ok = win32reg:change_key(Reg, "\\hklm"),
    ?line ok = win32reg:change_key(Reg, "software\\Ericsson\\Erlang"),
    ?line {ok, ErlangKey} = win32reg:current_key(Reg),
    io:format("Erlang key: ~s", [ErlangKey]),

    %% Get the values for the current Erlang version.
    %% Assume that the most recent comes last when the keys are sorted
    %% Check that the four standard values are part of the
    %% list returned.
    ?line {ok, Keys} = win32reg:sub_keys(Reg),
    ?line Current=erlang:info(version),
    ?line ok = win32reg:change_key(Reg, Current),
    ?line {ok, Values} = win32reg:values(Reg),
    io:format("Values for current: ~p", [Values]),
    ?line true = lists:keymember("Bindir", 1, Values),
    ?line true = lists:keymember("Emulator", 1, Values),
    ?line true = lists:keymember("Progname", 1, Values),
    ?line true = lists:keymember("Rootdir", 1, Values),

    %% Test relative navigation -- first upwards.

    ?line {ok, Here} = win32reg:current_key(Reg),
    ?line ok = win32reg:change_key(Reg, ".."),
    ?line {ok, ErlangKey} = win32reg:current_key(Reg),
    ?line {ok, ErlValues} = win32reg:values(Reg),
    
    %% Test relative navigation -- down again.

    ?line ok = win32reg:change_key(Reg, Current),
    ?line {ok, Here} = win32reg:current_key(Reg),
    ?line {ok, Values} = win32reg:values(Reg),
    
    %% Test relative navigation -- up and down in one step.

    ?line ok = win32reg:change_key(Reg, "..\\" ++ Current),
    ?line {ok, Here} = win32reg:current_key(Reg),
    ?line {ok, Values} = win32reg:values(Reg),

    %% Absolute navigation from the top.

    ?line ok = win32reg:change_key(Reg, "\\hkcr"),	% Get lost!
    ?line {ok, "\\hkey_classes_root"} = win32reg:current_key(Reg),
    ?line ok = win32reg:change_key(Reg, Here),
    ?line {ok, Here} = win32reg:current_key(Reg),
    ?line {ok, Values} = win32reg:values(Reg),
    
    %% Done.

    ?line ok = win32reg:close(Reg),
    ?line test_server:timetrap_cancel(Dog),
    ok.

long(doc) -> "Test long keys and entries (OTP-3446).";
long(Config) when list(Config) ->
    ?line Dog = test_server:timetrap(test_server:seconds(10)),

    ?line LongKey = "software\\" ++
	lists:flatten(lists:duplicate(10, "..\\software\\")) ++
	"Ericsson\\Erlang",
    ?line {ok,Reg} = win32reg:open([read,write]),
    ?line ok = win32reg:change_key(Reg, "\\hklm"),
    ?line ok = win32reg:change_key(Reg, LongKey),
    ?line {ok,ErlangKey} = win32reg:current_key(Reg),
    io:format("Erlang key: ~s", [ErlangKey]),

    %% Write a long value and read it back.
    ?line TestKey = "test_key",
    ?line LongValue = lists:concat(["This is a long value generated by the test case ",?MODULE,":long/1. "|lists:duplicate(128, "a")]),
    ?line ok = win32reg:set_value(Reg, TestKey, LongValue),
    ?line {ok,LongValue} = win32reg:value(Reg, TestKey),

    %% Done.

    ?line ok = win32reg:close(Reg),
    ?line test_server:timetrap_cancel(Dog),
    ok.

evil_write(Config) when list(Config) ->
    ?line Dog = test_server:timetrap(test_server:seconds(10)),

    ?line Key = "Software\\Ericsson\\Erlang",
    ?line {ok,Reg} = win32reg:open([read,write]),
    ?line ok = win32reg:change_key(Reg, "\\hklm"),
    ?line ok = win32reg:change_key(Reg, Key),
    ?line {ok,ErlangKey} = win32reg:current_key(Reg),
    io:format("Erlang key: ~s", [ErlangKey]),

    %% Write keys with different length and read it back.
    ?line TestKey = "test_key " ++ lists:duplicate(128, $a),
    evil_write_1(Reg, TestKey),

    %% Done.
    ?line ok = win32reg:close(Reg),
    ?line test_server:timetrap_cancel(Dog),
    ok.

evil_write_1(Reg, [_|[_|_]=Key]=Key0) ->
    ?line io:format("Key = ~p\n", [Key0]),
    ?line ok = win32reg:set_value(Reg, Key0, "A good value for me"),
    ?line {ok,Val} = win32reg:value(Reg, Key0),
    ?line ok = win32reg:delete_value(Reg, Key0),
    evil_write_1(Reg, Key);
evil_write_1(_, [_]) -> ok.
